from pathlib import Path

from pinjected import Design, instances, injected_function, Injected
import openai

__meta_design__: Design = instances(
    default_design_paths=["pinjected.demo.default_design"]
)

default_design: Design = instances(
    openai_api_key="my secret key",
    model="text-davinci-003",
    max_tokens=1000,
    load_openai_api_key=lambda: Path("~/openai_api_key.txt").expanduser().read_text().strip()
)


@injected_function
def LLM(load_openai_api_key: str, model, max_tokens, /, prompt) -> str:
    # openai_api_key = Path("~/openai_api_key.txt").expanduser().read_text().strip()
    return openai.Completion.create(
        load_openai_api_key(), prompt=prompt, model=model, max_tokens=max_tokens
    )["choices"][0]["text"]


@injected_function
def Gandalf(LLM: "str -> str", /, user_message: str) -> str:
    return LLM(f"Respond to the user's message as if you were a Gandalf: {user_message}")


test_greeting: Injected = Gandalf("How are you?")




from dataclasses import dataclass
from typing import TypeVar, Any, Callable, Generic, Iterator

from pinjected.di.proxiable import IProxyContext, DelegatedVar

T = TypeVar("T")
U = TypeVar("U")


@dataclass
class DynamicProxyContextImpl(IProxyContext[T]):
    accessor: Callable[[T], Any]
    pure_impl: Callable[[Any], T]
    _alias_name: str

    def getattr(self, tgt: T, name: str):
        return self.pure(getattr(self.accessor(tgt), name))

    def call(self, tgt: T, *args, **kwargs):
        return self.pure(self.accessor(tgt)(*args, **kwargs))

    def pure(self, tgt: T) -> DelegatedVar[T]:
        return DelegatedVar(self.pure_impl(tgt), self)

    def getitem(self, tgt: T, key) -> Any:
        return self.pure(self.accessor(tgt)[key])

    def eval(self, tgt: T):
        return self.accessor(tgt)

    def alias_name(self):
        return self._alias_name

    def iter(self, tgt: T):
        return DynamicProxyIterator(iter(self.accessor(tgt)), self)

    def dir(self, tgt: T):
        return dir(self.accessor(tgt))

    def map(self, tgt: T, f):
        return self.pure(f(self.accessor(tgt)))


@dataclass
class DynamicProxyIterator(Generic[T]):
    src: Iterator
    cxt: DynamicProxyContextImpl

    def __iter__(self):
        return self

    def __next__(self):
        return self.cxt.pure(next(self.src))



from dataclasses import dataclass

from pinjected import Designed, Injected, Design
from pinjected.di.applicative import Applicative
from pinjected.di.designed import PureDesigned
from pinjected.di.injected import InjectedPure
from pinjected.di.proxiable import T, DelegatedVar
from pinjected.di.static_proxy import eval_applicative, ast_proxy, AstProxyContextImpl
from pinjected.di.ast import Expr, Object, show_expr


class ApplicativeDesignedImpl(Applicative[Designed]):

    def map(self, target: Designed, f) -> T:
        return target.map(f)

    def zip(self, *targets: Designed):
        return Designed.zip(*targets)

    def pure(self, item) -> T:
        return Designed.bind(Injected.pure(item))

    def is_instance(self, item) ->bool:
        return isinstance(item,Designed)


def reduce_designed_expr(expr: Expr):
    match expr:
        case Object(PureDesigned(design, InjectedPure(value))):
            return f"Object({str(value)} with {design})"


@dataclass
class EvaledDesigned(Designed[T]):
    value: Designed[T]
    ast: Expr[Designed[T]]

    @property
    def design(self) -> "Design":
        return self.value.design

    @property
    def internal_injected(self) -> "Injected":
        return self.value.internal_injected

    def __str__(self):
        return f"EvaledDesigned(value={self.value},ast={show_expr(self.ast, reduce_designed_expr)})"


def eval_designed(expr: Expr[Designed]) -> Designed:
    return EvaledDesigned(eval_applicative(expr, ApplicativeDesigned), expr)


def designed_proxy(designed: Designed) -> DelegatedVar[Designed]:
    return ast_proxy(designed, DesignedEvalContext)


ApplicativeDesigned = ApplicativeDesignedImpl()
DesignedEvalContext = AstProxyContextImpl(eval_designed, _alias_name="DesignedProxy")



import ast
import inspect
import textwrap
from pathlib import Path
from pprint import pformat
from types import FrameType
from typing import TypeVar, Dict, Union

import cloudpickle
from cytoolz import memoize
from makefun import create_function
from returns.maybe import Some
from returns.result import Failure, Success

from pinjected.di.design import Design
from pinjected.di.bindings import InjectedBind, BindMetadata
from pinjected.di.injected import Injected, InjectedPure, InjectedFunction
from pinjected.di.metadata.location_data import CodeLocation, ModuleVarLocation
from pinjected.di.monadic import getitem_opt
from pinjected.di.proxiable import DelegatedVar

# is it possible to create a binding class which also has an ability to dynamically add..?
# yes. actually.
T = TypeVar("T")


def rec_valmap(f, tgt: dict):
    res = dict()
    for k, v in tgt.items():
        if isinstance(v, dict):
            res[k] = rec_valmap(f, v)
        else:
            res[k] = f(v)
    return res


def rec_val_filter(f, tgt: dict):
    res = dict()
    for k, v in tgt.items():
        if isinstance(v, dict):
            res[k] = rec_val_filter(f, v)
        elif f(v):
            res[k] = v
    return res


class ErrorWithTrace(BaseException):
    def __init__(self, src: BaseException, trace: str):
        super().__init__()
        self.src = src
        self.trace = trace

    def __reduce__(self):
        return (ErrorWithTrace, (self.src, self.trace))

    def __str__(self):
        return f"{self.src}\n {self.trace}"


def my_safe(f):
    def impl(*args, **kwargs):
        try:
            return Success(f(*args, **kwargs))
        except Exception as e:
            import traceback
            trace = "\n".join(traceback.format_exception(e))

            return Failure(ErrorWithTrace(
                e,
                trace
            ))

    return impl


def check_picklable(tgt: dict):
    cloud_dumps_try = my_safe(cloudpickle.dumps)
    cloud_loads_try = my_safe(cloudpickle.loads)
    res = cloud_dumps_try(tgt).bind(cloud_loads_try)

    if isinstance(res, Failure):
        # target_check = valmap(cloud_dumps_try, tgt)
        rec_check = rec_valmap(lambda v: (cloud_dumps_try(v), v), tgt)
        failures = rec_val_filter(lambda v: isinstance(v[0], Failure), rec_check)
        # failures = [(k, v, tgt[k]) for k, v in target_check.items() if isinstance(v, Failure)]

        from loguru import logger
        logger.error(f"Failed to pickle target: {pformat(failures)}")
        logger.error(f"if the error message contains EncodedFile pickling error, "
                     f"check whether the logging module is included in the target object or not.")
        raise RuntimeError("this object is not picklable. check the error messages above.") from res.failure()
    # logger.info(res)


def method_to_function(method):
    """
    converts a class method to a function
    """
    argspec = inspect.getfullargspec(method)
    assert not isinstance(argspec.args, str)
    # assert not isinstance(argspec.varargs,str)
    signature = f"""f_of_{method.__name__}({" ,".join((argspec.args or []))})"""

    def impl(self, *args, **kwargs):
        return method(*args, **kwargs)  # gets multiple values for self

    assert len(argspec.kwonlyargs) == 0, "providing method cannot have any kwonly args"
    return create_function(signature, impl)


def none_provider(func):
    argspec = inspect.getfullargspec(func)
    signature = f"""none_provider({" ,".join((argspec.args or []) + (argspec.varargs or []))})"""

    def impl(*args, **kwargs):
        func(*args, **kwargs)  # gets multiple values for self
        return "success of none_provider"

    assert len(argspec.kwonlyargs) == 0, "providing method cannot have any kwonly args"
    return create_function(signature, impl)


def extract_argnames(func):
    spec = inspect.getfullargspec(func)
    return spec.args


def get_class_aware_args(f):
    args = inspect.getfullargspec(f).args
    if isinstance(f, type) and "self" in args:
        args.remove("self")
    return args


def to_readable_name(o):
    match o:
        case InjectedBind(InjectedFunction(func, _)):
            return func.__name__
        case InjectedBind(InjectedPure(value)):
            return value
        case any:
            return any


def try_import_subject():
    try:
        from rx.subject import Subject
        return Subject
    except Exception as e:
        from rx.subjects import Subject
        return Subject


def get_dict_diff(a: dict, b: dict):
    all_keys = list(sorted(set(a.keys()) | set(b.keys())))
    all_keys.remove("opt")
    # TODO check both contains transform design
    # all_keys.remove("base_train_transform_design")
    # all_keys.remove("base_test_transform_design")
    all_keys.remove("design")
    data = []
    Subject = try_import_subject()
    for k in all_keys:

        ak = getitem_opt(a, k).map(to_readable_name).value_or(None)
        bk = getitem_opt(b, k).map(to_readable_name).value_or(None)
        match (ak, bk):
            case (Subject(), Subject()):
                flag = True
            case (a, b):
                flag = a != b
            case _:
                raise RuntimeError("this should not happen")
        if flag:
            data.append((k, ak, bk))

    return data


T = TypeVar("T")

EmptyDesign = Design()


# mapping is another layer of complication.
# good way is to create a MappedDesign class which is not a subclass of a Design
# only MappedDesign or Design can be __add__ ed to this class
# calling to_design converts all lazy mapping into providers
# so if anything is missing then fails.

def _get_external_type_name(thing):
    """patch pinject's _get_external_type_name to accept pickled function"""
    qualifier = thing.__qualname__
    name = qualifier.rsplit('.', 1)[0]
    if hasattr(inspect.getmodule(thing), name):
        cls = getattr(inspect.getmodule(thing), name)
        if isinstance(cls, type):
            return cls.__name__

    res = inspect.getmodule(thing)  # .__name__
    if res is None:
        return "unknown_module"
    return res.__name__


def instances(**kwargs):
    """
    Registers multiple concrete instances or values to a new ``Design`` instance and adds their respective code locations.

    This function is utilized for binding actual instances, as opposed to providers or factories, directly into a ``Design``. This is particularly useful for constants or already-constructed objects that you wish to make available for dependency injection. Each instance's code location is automatically registered, aiding in debugging and system introspection by maintaining information about where each instance was defined.

    The function performs assertions to ensure that no ``DelegatedVar`` or ``Injected`` instances are passed, preventing common errors.

    Parameters:
    -----------
    \*\*kwargs : dict
        Arbitrary keyword arguments representing the instances to be bound.
        Each key represents the name of the dependency, and the value is the corresponding concrete instance or value.
        The function asserts that none of the values are instances of ``DelegatedVar`` or ``Injected``,
        raising an assertion error if this case is encountered to prevent misuse.

    Returns:
    --------
    Design
        A ``Design`` instance with the concrete objects or values bound and their code locations registered.

    Raises:
    -------
    AssertionError
        If any value in ``kwargs`` is an instance of ``DelegatedVar`` or ``Injected``, an assertion error is raised,
        indicating that such usage is forbidden to prevent errors in binding.

    Example:
    --------
    .. code-block:: python

        # Pre-constructed objects or constants
        database_connection = create_database_connection()
        constant_value = 42

        # Create a design with instances registered
        design_with_instances = instances(
            database_connection=database_connection,
            constant=constant_value
        )

        # The 'design_with_instances' now has 'database_connection' and 'constant_value'
        # registered as concrete instances, along with their code locations.

    By enforcing constraints and adding code locations, this function helps maintain a clear and error-resistant
    system design, ensuring that the dependency injection mechanism is used appropriately and effectively.
    """
    for k, v in kwargs.items():
        assert not isinstance(v,
                              DelegatedVar), f"passing delegated var with Injected context is forbidden, to prevent human error."
        assert not isinstance(v,
                              Injected), f"key {k} is an instance of 'Injected'. passing Injected to 'instances' is forbidden, to prevent human error. use bind_instance instead."

    d = Design().bind_instance(**kwargs)
    return add_code_locations(d, kwargs, inspect.currentframe())


def providers(**kwargs):
    """
    Register multiple providers or `Injected` instances to a new `Design` instance, automatically adding code locations.

    This utility function simplifies the process of creating a `Design` instance and binding providers to it. It accepts both regular functions and instances of `Injected` as providers. These functions can be ones that require dependencies for their computations or `Injected` objects that encapsulate the dependency requirements.

    Each provider is automatically associated with its code location, aiding in system introspection and debugging by keeping track of where each provider was defined.

    Parameters:
    -----------
    \*\*kwargs : dict
        Arbitrary keyword arguments. Each key represents the name of the dependency, and the value can be either:
        1. A provider function, which is a callable that returns the necessary object for the dependency. This function can itself require other dependencies.
        2. An `Injected` instance which encapsulates both the creation of the dependency and its own dependencies.

    Returns:
    --------
    Design
        A `Design` instance with the providers or `Injected` instances bound and their code locations registered.

    Example:
    --------
    .. code-block:: python

        from pinjected.di.util import Injected

        def provide_database_connection():
            # logic to initiate a database connection
            pass

        injected_query_executor = Injected.bind(lambda connection: connection.execute_query())

        # Create a design with various types of providers
        design_with_providers = providers(
            database_connection=provide_database_connection,
            query_executor=injected_query_executor
        )

        # The 'design_with_providers' has the 'provide_database_connection' function and
        # 'injected_query_executor' Injected instance registered as providers for 'database_connection'
        # and 'query_executor', respectively, along with their code locations.

    This approach consolidates the registration of dependencies, streamlining the setup process and enhancing the traceability and debuggability of the system components.
    """
    d = Design().bind_provider(**kwargs)
    return add_code_locations(d, kwargs, inspect.currentframe())


def add_code_locations(design, kwargs, frame):
    locs = get_code_locations(list(kwargs.keys()), frame)
    metas = {k: BindMetadata(Some(loc)) for k, loc in locs.items()}
    return design.add_metadata(**metas)


@memoize
def try_parse(source: str, trials: int = 3) -> ast.AST:
    """
    Attempt to parse the source code, dedenting it if necessary.

    Args:
    - source (str): The source code to be parsed.
    - trials (int, optional): The maximum number of dedent trials. Defaults to 3.

    Returns:
    - ast.AST: The parsed AST of the source code.
    """
    try:
        return ast.parse(source)
    except IndentationError:
        if trials <= 0:
            raise  # re-raise the last IndentationError if max trials are exhausted
        return try_parse(textwrap.dedent(source), trials - 1)


def get_code_locations(keys: list[str], frame: FrameType) -> Dict[str, CodeLocation]:
    parent_frame = frame.f_back
    lines, start_line = inspect.getsourcelines(parent_frame)
    source = ''.join(lines)
    # logger.info(f"parsing:{source}")
    node = try_parse(source)

    locations = {}

    class ArgumentVisitor(ast.NodeVisitor):
        def visit_Call(self, node: ast.Call):
            node_lineno = node.lineno + start_line
            # parent_frame_lineno = parent_frame.f_lineno
            # print(f'node_match? {node_lineno} == {parent_frame_lineno} ?')
            # print(f"visit_Call:{node.lineno},{node.col_offset},{node},{start_line}, {parent_frame.f_lineno}")
            # print(f"node keywords:{node.keywords}")
            if node_lineno == parent_frame.f_lineno:  # The specific call that initiated the child frame
                for keyword in node.keywords:
                    if keyword.arg in keys:
                        locations[keyword.arg] = ModuleVarLocation(
                            Path(parent_frame.f_code.co_filename),
                            keyword.lineno + start_line,
                            keyword.col_offset + 1,
                        )

    ArgumentVisitor().visit(node)
    # print(locations)
    return locations


def get_code_location(frame):
    return ModuleVarLocation(
        Path(frame.f_code.co_filename),
        frame.f_lineno,
        0,
    )


def classes(**kwargs):
    """
    Binds classes to a new ``Design`` instance, adding code locations for reference.

    This function is a utility for binding classes to a new design, making them available for dependency injection. It also registers their code locations, aiding in debugging by preserving the origin information of each class.

    Parameters:
    -----------
    \*\*kwargs : dict
        Arbitrary keyword arguments. Each key is a string that represents the dependency name, and the value is the class to be bound.

    Returns:
    --------
    Design
        Returns a new ``Design`` instance with the specified classes bound.

    Example:
    --------
    .. code-block:: python

        class MyClass:
            pass

        # Binding the class within a new design
        design_with_class = classes(MyClass=MyClass)

        # 'design_with_class' now has 'MyClass' available for dependency injection.

    Note:
    -----
    The function automatically adds code locations for the classes, making it easier to track the source of each dependency within the system.
    """
    d = Design().bind_class(**kwargs)
    return add_code_locations(d, kwargs, inspect.currentframe())


def injecteds(**kwargs):
    return Design().bind_provider(**kwargs)



from returns.result import safe, Failure

from pinjected import Design


def test_provide_session():
    d = Design().bind_instance(
        a=0
    )
    g = d.to_graph()
    assert g["session"] == g


def test_child_session():
    d = Design().bind_instance(
        a=0
    ).bind_provider(
        x=lambda a: a
    )
    g = d.to_graph()
    child_g = g.child_session(Design().bind_instance(a=1))
    grandchild_g = child_g.child_session(Design().bind_instance(a=2))
    assert g["a"] == 0
    assert child_g['a'] == 1  # ah, this is actually an expected behavior.
    assert grandchild_g['a'] == 2
    assert child_g['a'] == 1  # oh why is this 0??
    assert g["a"] == 0
    assert g["x"] == 0
    assert child_g['x'] == 0  # x is already in parent, and is not overriden explicitly.
    # for this to work we need to track all the dependenciy tree of a binding.
    assert grandchild_g['x'] == 0
    assert child_g['x'] == 0  # oh why is this 0??
    assert g["x"] == 0
    # so in order to make session work,
    # you must specify all the components,
    # implicit components in child session will be forgotten.


def test_proxied():
    d= Design().bind_instance(
    ).bind_provider(
        y = lambda x:x+1
    )
    g = d.to_graph()
    evaluated = g.proxied("y").eval()
    print(evaluated)
    assert isinstance(safe(evaluated.run)(),Failure)
    assert evaluated.override(Design().bind_instance(x=1)).run() == 2
    assert evaluated.override(Design().bind_instance(x=2)).run() == 3



from pinjected import Injected, injected, instance
from pinjected.di.util import instances, providers

__meta_design__ = instances(
    name="test_package.child.module1",
    default_design_paths=[
        "pinjected.test_package.child.module1.design01"
    ]
)


design01 = instances(name='design01')
design02 = design01 + instances(name='design02')
a = Injected.pure('a')
b = Injected.pure('b')
@instance
def test_viz_target(a,b):
    return a + b
viz_target_design = providers(
    a=a,
    b=b
)
test_runnable = Injected.pure("hello world")



from dataclasses import dataclass
from loguru import logger
from pathlib import Path
from typing import Dict, List

from pinjected import Design, Injected, instances
from pinjected.module_helper import walk_module_attr
from pinjected.module_inspector import ModuleVarSpec


@dataclass
class IdeaRunConfiguration:
    name: str
    script_path: str
    interpreter_path: str
    arguments: List[str]
    working_dir: str


@dataclass
class IdeaRunConfigurations:
    configs: Dict[str, List[IdeaRunConfiguration]]


@dataclass
class MetaContext:
    trace: List[ModuleVarSpec[Design]]
    accumulated: Design

    @staticmethod
    def gather_from_path(file_path: Path, meta_design_name: str = "__meta_design__"):
        if not isinstance(file_path, Path):
            file_path = Path(file_path)
        designs = list(walk_module_attr(file_path, meta_design_name))
        designs.reverse()
        res = Design()
        overrides = Design()
        for item in designs:
            logger.debug(f"{meta_design_name} at :{item.var_path}")
            res = res + item.var
            try:
                overrides += item.var.provide("overrides")
            except Exception as e:
                logger.debug(f"{item.var_path} does not contain overrides")
        from pinjected import instances
        res += instances(
            overrides=overrides
        )

        return MetaContext(
            trace=designs,
            accumulated=res
        )


@dataclass
class RunnablePair:
    target: Injected
    design: Design

    def run(self):
        logger.info(f"running {self.target}")
        result = self.design.to_graph()[self.target]
        logger.info(f"result: {result}")
        return result

    def save_html(self, name: str = None, show=True):
        if name is None:
            name = "graph.html"
        self.design.to_vis_graph().save_as_html(self.target, name, show=show)


try:
    # pydantic over version 2
    from pydantic import field_validator
except ImportError:
    from pydantic import validator as field_validator

__meta_design__ = instances(
    default_design_paths=["pinjected.helper_structure.__meta_design__"]
)



from expression import Nothing
from returns import maybe as raybe
from returns.maybe import Some, Maybe


def patch_maybe():
    def maybe__or__(self, other):
        match self:
            case Some(x):
                return self
            case raybe.Nothing:
                return other

    def maybe_filter(self: Maybe, flag_to_keep):
        match self.map(flag_to_keep):
            case Some(True):
                return self
            case _:
                return Nothing

    Maybe.__or__ = maybe__or__
    Maybe.filter = maybe_filter



from contextlib import contextmanager
from loguru import logger


@contextmanager
def disable_internal_logging():
    # logger.info(f"disabling internal logging")
    names = [
        'pinjected.di.graph',
        'pinjected.helpers',
        'pinjected.module_inspector',
        'pinjected'
    ]
    for n in names:
        logger.disable(n)
    yield
    for n in names:
        logger.enable(n)
    # logger.info(f"enabling internal logging")



from pinjected import Injected, Designed
from pinjected.di.static_proxy import ast_proxy
from pinjected.di.ast import Object


def test_expr():
    print(Object("x").test.x.y[0])


def test_ast_proxy():
    print(ast_proxy("hello").test.x.y.z[0].eval())


def test_eval_injected():
    tgt = Injected.pure("hello").proxy.x.y.z
    a = Injected.pure("a").proxy
    b = Injected.pure("b").proxy
    ast = tgt(a, b=b)
    print(ast)
    print(ast.eval())

def test_designed():

    tgt = Designed.bind(Injected.pure("hello")).proxy.x.y.z
    a = Designed.bind(Injected.pure("a")).proxy
    b = Designed.bind(Injected.pure("b")).proxy
    ast = tgt(a, b=b)
    print(ast)
    print(ast.eval())



from dataclasses import dataclass

from pinjected import Design, Designed, Injected, EmptyDesign
from pinjected.di.dynamic_proxy import DynamicProxyContextImpl
from pinjected.di.graph import SessionValue


@dataclass
class A:
    item: str


def test_dynamic_proxy_iterator():
    ctx = DynamicProxyContextImpl(
        lambda a: a.item,
        A,
        "A_Proxy"
    )
    x = ctx.pure("hello world")
    print(x)
    print(x.split())
    print([i for i in x.split()])


def test_session_value_iterator():
    d = Design().bind_instance(x=0)
    g = d.to_graph()
    session = g.child_session(EmptyDesign)
    ctx = DynamicProxyContextImpl(
        lambda a: a.value,
        lambda x: SessionValue(
            g,
            Designed.bind(Injected.pure(x)),
            session,
        ),
        "SessionValueProxy"
    )
    x = ctx.pure("hello world")
    print(x)
    print(x.split())
    for item in x.split():
        print(item.eval())

    print(g.sessioned("x"))




INSPECT_TEMPLATE = f"""
There is a function annotatation called 'injected' that converts a function into a wrapped function.
Example:
@injected
def foo(dep1,dep2,/,x:int) -> int:
    return 1

foo: InjectedProxy[int -> int]
foo_call_result: Injected[int] = foo(1)

Injected is a monad that represents a value which requires dependency injection.
In the example above, the positional only arguments are considered as dependencies and get injected by the framework.
As a result, the function ends up with a single argument x:int and returns an int.
In the example above, the foo function is wrapped by InjectedProxy, which acts as a proxy for the call or attribute access inside the Injected monad. 
"""



from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Generic, Tuple, Dict, Any, Callable, Optional, TypeVar

from frozendict import frozendict

from pinjected.di.injected_analysis import get_instance_origin
from pinjected.global_configs import pinjected_TRACK_ORIGIN

T = TypeVar("T")


class Expr(Generic[T], ABC):
    """
    I want to know where this instance is instantiated
    """

    def __post_init__(self):
        # this takes time so... we need to toggle it
        if pinjected_TRACK_ORIGIN:
            self.origin_frame = get_instance_origin('pinjected')
        else:
            self.origin_frame = None

    def __getattr__(self, item: str):
        return Attr(self, item)

    def _wrap_if_non_expr(self, item) -> "Expr":
        if not isinstance(item, Expr):
            return Object(item)
        else:
            return item

    def __call__(self, *args: "Expr", **kwargs: "Expr"):
        # print(f"{self}->args:{args},kwargs:{kwargs}")
        return Call(self,
                    tuple([self._wrap_if_non_expr(item) for item in args]),
                    {k: self._wrap_if_non_expr(v) for k, v in kwargs.items()}
                    )

    def __getitem__(self, item: "Expr"):
        return GetItem(self, self._wrap_if_non_expr(item))

    @abstractmethod
    def __getstate__(self):
        pass

    @abstractmethod
    def __setstate__(self, state):
        pass

    def __str__(self):
        return f"Expr(>{show_expr(self)}<)"

    def __repr__(self):
        return str(self)

    def __add__(self, other):
        return BiOp("+", self, self._wrap_if_non_expr(other))

    def __or__(self, other):
        return BiOp("|", self, self._wrap_if_non_expr(other))

    def __and__(self, other):
        return BiOp("&", self, self._wrap_if_non_expr(other))

    def __sub__(self, other):
        return BiOp("-", self, self._wrap_if_non_expr(other))

    def __mul__(self, other):
        return BiOp("*", self, self._wrap_if_non_expr(other))

    def __matmul__(self, other):
        return BiOp("@", self, self._wrap_if_non_expr(other))

    def __truediv__(self, other):
        return BiOp("/", self, self._wrap_if_non_expr(other))

    def __floordiv__(self, other):
        return BiOp("//", self, self._wrap_if_non_expr(other))

    def __mod__(self, other):
        return BiOp("%", self, self._wrap_if_non_expr(other))

    def __pow__(self, other):
        return BiOp("**", self, self._wrap_if_non_expr(other))

    def __lshift__(self, other):
        return BiOp("<<", self, self._wrap_if_non_expr(other))

    def __rshift__(self, other):
        return BiOp(">>", self, self._wrap_if_non_expr(other))

    def __xor__(self, other):
        return BiOp("^", self, self._wrap_if_non_expr(other))

    def __eq__(self, other):
        return BiOp("==", self, self._wrap_if_non_expr(other))


@dataclass
class BiOp(Expr):
    name: str
    left: Expr
    right: Expr

    def __post_init__(self):
        super().__post_init__()
        assert isinstance(self.left, Expr), f"{self.left} is not an Expr"
        assert isinstance(self.right, Expr), f"{self.right} is not an Expr"
        assert isinstance(self.name, str), f"{self.name} is not a str"

    def __getstate__(self):
        return self.name, self.left, self.right, self.origin_frame

    def __setstate__(self, state):
        self.name, self.left, self.right, self.origin_frame = state

    def __hash__(self):
        return hash((self.name, self.left, self.right))


@dataclass
class Call(Expr):
    func: Expr
    args: Tuple[Expr] = field(default_factory=tuple)
    kwargs: Dict[str, Expr] = field(default_factory=dict)

    def __getstate__(self):
        return self.func, self.args, self.kwargs, self.origin_frame

    def __setstate__(self, state):
        self.func, self.args, self.kwargs, self.origin_frame = state

    def __hash__(self):
        return hash((self.func, self.args, frozendict(self.kwargs)))


@dataclass
class Attr(Expr):
    data: Expr
    attr_name: str  # static access so no ast involved

    def __getstate__(self):
        return self.data, self.attr_name, self.origin_frame

    def __setstate__(self, state):
        self.data, self.attr_name, self.origin_frame = state

    def __hash__(self):
        return hash((self.data, self.attr_name))


@dataclass
class GetItem(Expr):
    data: Expr
    key: Expr

    def __getstate__(self):
        return self.data, self.key, self.origin_frame

    def __setstate__(self, state):
        self.data, self.key, self.origin_frame = state

    def __hash__(self):
        return hash((self.data, self.key))


@dataclass
class Object(Expr):
    """
    Use this to construct an AST and then compile it for any use.
    """
    data: Any  # holds user data

    def __getstate__(self):
        return self.data, self.origin_frame

    def __setstate__(self, state):
        self.data, self.origin_frame = state

    def __hash__(self):
        # what if the data is not hashable?
        try:
            return hash(self.data)
        except TypeError as e:
            return hash(id(self.data))


def show_expr(expr: Expr[T], custom: Callable[[Expr[T]], Optional[str]] = lambda x: None) -> str:
    from pinjected.di.proxiable import DelegatedVar
    def _show_expr(expr):
        def eval_tuple(expr):
            return tuple(_show_expr(i) for i in expr)

        def eval_dict(expr):
            return {k: _show_expr(e) for k, e in expr.items()}

        reduced = custom(expr)
        if reduced:
            return reduced
        match expr:
            case Object(str() as x):
                return f'"{x}"'
            case Object(x):
                return str(x)
            case Call(f, args, kwargs):
                # hmm, this is complicated.
                func_str = _show_expr(f)
                args = list(eval_tuple(args))
                kwargs = eval_dict(kwargs)
                kwargs = [f"{k}={v}" for k, v in kwargs.items()]
                return f"{func_str}({','.join(args + kwargs)})"
            case BiOp(str() as op, Expr() as left, Expr() as right):
                return f"({_show_expr(left)} {op} {_show_expr(right)})"
            case Attr(Expr() as data, str() as attr_name):
                return f"{_show_expr(data)}.{attr_name}"
            case GetItem(Expr() as data, key):
                return f"{_show_expr(data)}[{_show_expr(key)}]"
            case DelegatedVar(wrapped, cxt):
                return f"{_show_expr(wrapped)}"

            case _:
                raise RuntimeError(f"unsupported ast found!:{type(expr)}")

    return _show_expr(expr)



import inspect
from typing import Union, Callable

from returns.maybe import Some

from pinjected import Injected
from pinjected.di.implicit_globals import IMPLICIT_BINDINGS
from pinjected.di.injected import PartialInjectedFunction
from pinjected.di.util import get_code_location
import functools
import asyncio


def injected_function(f, parent_frame=None) -> PartialInjectedFunction:
    """
    Wraps a function, injecting dependencies for parameters that start with an underscore or are positional-only.
    This enhanced version supports class methods, recognizing and bypassing the 'self' parameter automatically.
    The function also registers the newly created function in the ``IMPLICIT_BINDINGS`` global dictionary.

    :param f: The target function where dependencies will be injected. This can be a standard function or a class method.
    :param parent_frame: The parent frame to inspect for getting the code location. If not provided, the system automatically determines the appropriate frame.
    :return: A new function with dependencies injected, suitable for both standalone functions and class methods.

    **Example Usage:**

    Given a class with a method that requires dependency injection:

    .. code-block:: python
        @injected_function
        class MyClass:
            dependency1: object
            def my_method(self, arg2):
                # Method implementation here
                # use self.dependency1

        # now we have Injected[Callable[[],MyClass]]
        g[MyClass]() # returns an instance of MyClass with dependency1 injected

    .. note::
        The function or a class will be automatically registered with its name, in the global ``IMPLICIT_BINDINGS`` dictionary, making it recognizable by the system's dependency injection mechanisms.
    """
    # How can we make this work on a class method?
    sig: inspect.Signature = inspect.signature(f)
    tgts = dict()
    if parent_frame is None:
        parent_frame = inspect.currentframe().f_back
    for k, v in sig.parameters.items():
        if k.startswith("_"):
            tgts[k] = Injected.by_name(k[1:])
        elif v.kind == inspect.Parameter.POSITIONAL_ONLY:
            tgts[k] = Injected.by_name(k)
    new_f = Injected.partial(f, **tgts)

    from pinjected.di.bindings import InjectedBind
    from pinjected.di.bindings import BindMetadata
    IMPLICIT_BINDINGS[f.__name__] = InjectedBind(
        new_f,
        metadata=Some(BindMetadata(code_location=Some(get_code_location(parent_frame)))),
    )

    return new_f


def injected_instance(f) -> Injected:
    """
    The ``injected_instance`` (also accessible as ``instance``) is a decorator that creates an
    ``Injected`` instance from a function. This function essentially converts a regular function
    that returns an instance into an ``Injected`` type that can be used for dependency injection,
    particularly within the context of a design setup.

    :param f: The function that generates an instance, typically utilizing dependencies.
    :type f: Function

    :return: An instance of ``Injected`` that wraps the provided function.
    :rtype: Injected

    :Example:

    .. code-block:: python

        @instance
        def logger(dep1, dep2):
            from logging import getLogger
            # maybe use dep1, dep2...
            return getLogger(__name__)

        # Now, 'logger' is an instance of 'Injected' specialized with 'Logger'.
        assert type(logger) == Injected  # True for Injected[Logger]

        # This 'Injected' instance can be contributed to a design like so:
        design = providers(
            logger=logger
        )

        # Usage of the designed provider
        _logger = design.provide('logger')  # returns a Logger instance
        _logger.info('hello world')

    This approach allows the function to be integrated into a dependency injection system easily,
    with the ``Injected`` instance handling the complexities of resolving and managing dependencies.
    """

    is_coroutine = inspect.iscoroutinefunction(f)
    if is_coroutine:
        f = cached_coroutine(f)

    sig: inspect.Signature = inspect.signature(f)
    tgts = {k: Injected.by_name(k) for k, v in sig.parameters.items()}
    instance = Injected.partial(f, **tgts)().eval()
    from pinjected.di.bindings import InjectedBind
    from pinjected.di.bindings import BindMetadata
    IMPLICIT_BINDINGS[f.__name__] = InjectedBind(
        instance,
        Some(BindMetadata(code_location=Some(get_code_location(inspect.currentframe().f_back))))
    )
    instance.__is_awaitable__ = is_coroutine
    return instance


def injected(tgt: Union[str, type, Callable]):
    """
    The ``injected`` decorator automates dependency injection, transforming a target (string, class, or callable) 
    into an ``Injected`` instance. It specifically treats positional-only parameters as dependencies, 
    automatically injecting them. In contrast, non-positional-only parameters are left as-is for later 
    specification during function or method invocation.

    :param tgt: The target indicating what is to be injected. This can be:
                1. ``str``: the name of a dependency.
                2. ``type``: a class that needs automated dependency injection for instantiation.
                3. ``Callable``: a function or method requiring dependencies.
    :return: An appropriate ``Injected`` instance, proxy, or wrapped entity with dependencies injected, 
             contingent on the nature of ``tgt``.

    **Usage Example:**

    .. code-block:: python

        @injected
        def function_requiring_dependencies(dependency1: Type1, /, normal_param: Type2):
            # Function body here. 'dependency1' is injected, 'normal_param' must be provided during call.

        @injected
        class ClassRequiringDependencies:
            def __init__(self, dependency1: Type1, /, normal_param: Type2):
                # Constructor body. 'dependency1' is injected, 'normal_param' is left for object creation time.

        # For direct dependency retrieval via a string identifier.
        dependency_instance = injected("dependency_key")

    In these examples, ``dependency1`` is a positional-only parameter and treated as a dependency to be 
    automatically injected. On the other hand, ``normal_param`` is a non-positional-only parameter. It's 
    not considered a dependency within the automatic injection process, and thus, must be specified 
    during the routine call or object instantiation.
    

    .. note::
        This approach enforces clear demarcation between automatically resolved dependencies 
        (positional-only) and those parameters that developers need to provide explicitly during 
        function/method invocation or class instantiation. This strategy enhances code readability 
        and ensures that the dependency injection framework adheres to explicit programming practices.
    """
    if isinstance(tgt, str):
        return Injected.by_name(tgt).proxy
    elif isinstance(tgt, type):
        return injected_function(tgt, parent_frame=inspect.currentframe().f_back)
    elif callable(tgt):
        return injected_function(tgt, parent_frame=inspect.currentframe().f_back)


def injected_class(cls):
    return injected_function(cls)


def injected_method(f):
    _impl = injected_function(f)

    def impl(self, *args, **kwargs):
        return _impl(self, *args, **kwargs)

    return impl


class CachedAwaitable:
    def __init__(self, coro):
        from loguru import logger
        #logger.warning(f'CachedAwaitable created with {coro}')
        self.coro = coro
        self._cache = None
        self._has_run = False
        self._lock = asyncio.Lock()

    def __await__(self):
        return self._get_result().__await__()

    async def _get_result(self):
        from loguru import logger
        #logger.warning(f"accessing cached coroutine:{self.coro}")
        async with self._lock:
            if not self._has_run:
                try:
                    self._cache = await self.coro
                except Exception as e:
                    logger.error(f"cached coroutine failed with {e},{self.coro}")
                    self._cache = e
                    raise e
                finally:
                    self._has_run = True
        if isinstance(self._cache, Exception):
            raise self._cache
        return self._cache


def cached_coroutine(coro_func):
    @functools.wraps(coro_func)
    def wrapper(*args, **kwargs):
        return CachedAwaitable(coro_func(*args, **kwargs))

    functools.update_wrapper(wrapper, coro_func)
    return wrapper


instance = injected_instance




import importlib
from dataclasses import dataclass, field
from pathlib import Path
from returns.maybe import maybe
from typing import Optional

from pinjected import injected_function
from pinjected.helper_structure import IdeaRunConfigurations
from pinjected.maybe_patch import patch_maybe
from pinjected.module_helper import walk_module_attr
from pinjected.module_inspector import ModuleVarSpec
from pinjected.module_var_path import ModuleVarPath
from pinjected.runnables import get_runnables

patch_maybe()

@injected_function
def inspect_and_make_configurations(
        injected_to_idea_configs,
        logger,
        /,
        module_path
) -> IdeaRunConfigurations:
    runnables = get_runnables(module_path)
    logger.info(f"Found {len(runnables)} injecteds")
    results = dict()
    logger.info(f"found {len(runnables)} injecteds")
    for tgt in runnables:
        if isinstance(tgt, ModuleVarSpec):
            results.update(injected_to_idea_configs(tgt).configs)
    return IdeaRunConfigurations(configs=results)


@dataclass
class RunnableSpec:
    tgt_path: ModuleVarPath
    design_path: ModuleVarPath = field(default_factory=lambda: ModuleVarPath("pinjected.di.util.EmptyDesign"))

    def __post_init__(self):
        # add type check
        assert isinstance(self.tgt_path, ModuleVarPath)
        assert isinstance(self.design_path, ModuleVarPath)

    @property
    def target_name(self):
        return self.tgt_path.var_name

    @property
    def design_name(self):
        return self.design_path.var_name


def get_design_path_from_var_path(var_path):
    from loguru import logger
    logger.info(f"looking for default design paths from {var_path}")
    module_path = ".".join(var_path.split(".")[:-1])
    # we need to get the file path from var path
    logger.info(f"loading module:{module_path}")
    # Import the module
    module = importlib.import_module(module_path)
    module_path = module.__file__
    design_paths = find_default_design_paths(module_path, None)
    #assert design_paths, f"no default design paths found for {var_path}"
    if design_paths:
        design_path = design_paths[0]
        return design_path


def find_default_design_path(file_path: str) -> Optional[str]:
    from loguru import logger
    logger.info(f"looking for default design_path")
    return find_module_attr(file_path, '__default_design_path__')


def find_default_working_dir(file_path: str) -> Optional[str]:
    from loguru import logger
    logger.info(f"looking for default working dir")
    return find_module_attr(file_path, '__default_working_dir__')


def find_default_design_paths(module_path, default_design_path: Optional[str]) -> list[str]:
    """
    :param module_path: absolute file path.("/")
    :param default_design_path: absolute module path (".")
    :return:
    """
    default_design_paths: list[str] = maybe(find_module_attr)(module_path, '__default_design_paths__').value_or([])
    default_design_path: list[str] = (
            maybe(lambda: default_design_path)() | maybe(find_default_design_path)(module_path)).map(
        lambda p: [p]).value_or([])
    from loguru import logger
    logger.debug(f"default design paths:{default_design_paths}")
    logger.debug(f"default design path:{default_design_path}")
    default_design_paths = default_design_paths + default_design_path
    return default_design_paths


def find_module_attr(file_path: str, attr_name: str, root_module_path: str = None) -> Optional[str]:
    for item in walk_module_attr(Path(file_path), attr_name, root_module_path):
        return item.var





import inspect
from copy import copy
from dataclasses import dataclass, field, replace
from functools import wraps
from itertools import chain
from typing import TypeVar, List, Dict, Union, Callable, Type

from cytoolz import merge
from makefun import create_function
from returns.maybe import Some

from pinjected.di.injected import Injected
from pinjected.di.bindings import Bind, InjectedBind, BindMetadata
from pinjected.di.graph import IObjectGraph, MyObjectGraph

from pinjected.di.injected import extract_dependency_including_self, InjectedPure, InjectedFunction
from pinjected.di.proxiable import DelegatedVar
# from pinjected.di.util import get_class_aware_args, get_dict_diff, check_picklable
from pinjected.di.monadic import getitem_opt
from pinjected.graph_inspection import DIGraphHelper

T = TypeVar("T")
U = TypeVar("U")


def map_wrap(src, f):
    @wraps(src)
    def wrapper(*args, **kwargs):
        return f(src(*args, **kwargs))

    wrapper.__signature__ = inspect.signature(src)
    return wrapper


def remove_kwargs_from_func(f, kwargs: List[str]):
    deps = extract_dependency_including_self(f)
    to_remove = set(kwargs)
    new_kwargs = deps - to_remove
    func_name = f.__name__.replace("<lambda>", "_lambda_")
    sig = f"""{func_name}({",".join(new_kwargs)})"""

    def impl(**called_kwargs):
        deleted = deps & to_remove
        # for self, you must check whether f is method or not
        if inspect.ismethod(f):
            deleted = deleted - {"self"}
        d_kwargs = {k: None for k in deleted}
        return f(**called_kwargs, **d_kwargs)

    return create_function(sig, impl)


@dataclass
class Design:
    """
    The ``Design`` class acts as a central registry for managing and applying dependency injection
    within your application. It allows for the binding of various components, instances, and providers,
    enabling a cohesive and flexible way to handle dependencies. The class is integral in constructing
    applications that adhere to the Dependency Inversion Principle, fostering a decoupled and easily
    maintainable codebase.

    Basic Usage:
    ------------

    **Adding Bindings:**

    The ``Design`` class allows for the binding of instances, providers, and classes.
    Each type of binding adds different kinds of objects to the design, contributing
    to the overall dependency structure.

    .. code-block:: python

        from pinjected import Design
        from dataclasses import dataclass

        @dataclass
        class DepObject:
            a: int
            b: int
            c: int
            d: int

        @dataclass
        class App:
            dep: DepObject

            def run(self):
                print(self.dep.a + self.dep.b + self.dep.c + self.dep.d)

        d = Design().bind_instance(
            a=0,
            b=1
        ).bind_provider(
            c=lambda a, b: a + b,
            d=lambda a, b, c: a + b + c
        ).bind_class(
            dep=DepObject
        )

        # The `to_graph` method compiles the bindings into a graph structure, which can then provide
        # the necessary components for your application.
        d.to_graph().provide(App).run()

    Advanced Usage:
    ---------------

    **Combining Multiple Designs:**

    In more complex scenarios, you might need to combine multiple ``Design`` instances.
    This could be necessary if different modules or components of your application require
    different dependency configurations. The ``Design`` class supports the combination of instances,
    allowing for the overlaying and overriding of bindings.

    .. code-block:: python

        d1 = Design().bind_instance(
            a=0
        )
        d2 = Design().bind_instance(
            b=1
        )
        d3 = Design().bind_instance(
            b=0
        )

        # The `+` operator combines designs. If the same binding exists in both designs,
        # the one from the rightmost design (latest) will take precedence.
        assert (d1 + d2).provide("b") == 1  # 'b' from d2 takes precedence
        assert (d1 + d2 + d3).provide("b") == 0  # 'b' from d3 overrides the others

    This feature ensures that ``Design`` instances are composable and adaptable, providing
    a robust foundation for building complex, modular applications with dependency injection.
    """
    bindings: Dict[str, Bind] = field(default_factory=dict)
    multi_binds: dict = field(default_factory=dict)
    modules: list = field(default_factory=list)
    classes: list = field(default_factory=list)

    def __getstate__(self):
        res = dict(
            bindings=self.bindings,
            multi_binds=self.multi_binds,
            modules=[m.__name__ for m in self.modules],
            classes=self.classes,
        )
        return res

    def __setstate__(self, state):
        mods = state["modules"]
        import importlib
        mods = [importlib.import_module(
            name=m
        ) for m in mods]
        state["modules"] = mods
        for k, v in state.items():
            setattr(self, k, v)

    def __add__(self, other: Union["Design", Dict[str, Bind]]):
        other = self._ensure_dbc(other)
        return self.merged(other)

    def bind(self, key: str) -> "DesignBindContext":

        return DesignBindContext(self, key)

    def _ensure_dbc(self, other: Union["Design", Dict[str, Bind]]):
        match other:
            case Design():
                return other
            case dict() if all([isinstance(v, Bind) for v in other.values()]):
                return Design(other)
            case _:
                raise ValueError(f"cannot add {type(other)} to Design")

    def _merge_multi_binds(self, src, dst):
        keys = src.keys() | dst.keys()
        multi = {k: (
                getitem_opt(src, k).value_or([]) +
                getitem_opt(dst, k).value_or([])
        ) for k in keys}
        return multi

    def merged(self, other: "Design"):
        """creates another instance with merged bindings. does not modify self"""
        # logger.debug(f"merging:\n\t{self} to \n\t{other}")
        assert isinstance(other, Design), f"merge target is not a Design. type:{type(other)}:{other}"

        res = Design(
            bindings=merge(self.bindings, other.bindings),
            multi_binds=self._merge_multi_binds(self.multi_binds, other.multi_binds),
            modules=list(set(self.modules) | set(other.modules)),
            classes=list(set(self.classes) | set(other.classes)),
        )
        return res

    def bind_instance(self, **kwargs, ):
        """
        Here, I need to find the CodeLocation for each binding.
        :param kwargs:
        :param __binding_metadata__: a dict of [str,BindMetadata]
        :return:
        """
        x = self
        for k, v in kwargs.items():
            if isinstance(v, type):
                from loguru import logger
                logger.warning(f"{k} is bound to class {v} with 'bind_instance' do you mean 'bind_class'?")
            x += Design({k: InjectedBind(InjectedPure(v))})
        return x

    def bind_provider(self, **kwargs: Union[Callable, Injected]):
        x = self
        for k, v in kwargs.items():
            # logger.info(f"binding provider:{k}=>{v}")
            def parse(item):
                from loguru import logger
                match item:
                    case type():
                        logger.warning(f"{k}->{v}: class is used for bind_provider. fixing automatically.")
                        return x.bind(k).to_class(item)
                    case Injected():
                        return x.bind(k).to_provider(item)
                    case DelegatedVar():
                        return parse(item.eval())
                    case non_func if not callable(non_func):
                        logger.warning(
                            f"{k}->{item}: non-callable or non-injected is passed to bind_provider. fixing automatically.")
                        return x.bind(k).to_instance(v)
                    case _:
                        return x.bind(k).to_provider(item)

            x = parse(v)
        return x

    def bind_class(self, **kwargs):
        from loguru import logger
        x = self
        for k, v in kwargs.items():
            if isinstance(v, Injected):
                logger.warning(f"{k}->{v}: Injected instance is used for bind_class. fixing automatically.")
                x = x.bind(k).to_provider(v)
            else:
                x = x.bind(k).to_class(v)
        return x

    def add_metadata(self, **kwargs: "BindMetadata") -> "Design":
        res = self
        for k, meta in kwargs.items():
            res += Design(
                bindings={k: InjectedBind(self[k].to_injected(), metadata=Some(meta))}
            )
        return res

    def to_graph(self, modules=None, classes=None, trace_logger=None) -> IObjectGraph:
        # So MyObjectGraph's session is still corrupt?
        # TODO add special variables to monitor the state of this provider.
        # __pinjected_events__ <= subject of events that are emitted by pinjected
        # but I don't want to depend on rx, right?
        design = self + Design(
            modules=modules or [],
            classes=classes or []
        )
        return MyObjectGraph.root(design, trace_logger=trace_logger)

    def run(self, f, modules=None, classes=None):
        return self.to_graph(modules, classes).run(f)

    def provide(self, target: Union[str, Type[T]], modules=None, classes=None, trace_logger=None) -> T:
        """
        :param target: provided name
        :param modules: modules to use for graph construction
        :return:
        """
        return self.to_graph(modules=modules, classes=classes, trace_logger=trace_logger).provide(target, level=4)

    def copy(self):
        return self.__class__(
            bindings=self.bindings.copy(),
            multi_binds=copy(self.multi_binds),
            modules=copy(self.modules),
            classes=copy(self.classes),
        )

    def map_value(self, src_key, f):
        """
        :param src_key:
        :param f:
        :return: Design
        """
        mapped_binding = self.bindings[src_key].map(f)
        return self + Design({src_key: mapped_binding})

    def apply_injected_func(self, key: str, injected_func: Injected[Callable]):
        bind = self.bindings[key]
        applied_bind = InjectedBind(
            bind.to_injected().apply_injected_function(injected_func),
        )
        return self + Design({key: applied_bind})

    def keys(self):
        return self.bindings.keys()

    def unbind(self, key) -> "Design":
        if key in self.bindings:
            copied = self.bindings.copy()
            del copied[key]
            return replace(self,
                           bindings=copied
                           )
        return self

    def __contains__(self, item):
        return item in self.bindings

    def __getitem__(self, item):
        return self.bindings[item]

    def __str__(self):
        return f"Design(len={len(self.bindings) + len(self.multi_binds)})"

    def __repr__(self):
        return str(self)

    def table_str(self):
        import tabulate
        binds = tabulate.tabulate(sorted(list(self.bindings.items())))
        multis = tabulate.tabulate(sorted(list(self.multi_binds.items())))
        return binds + "\n" + multis

    def to_str_dict(self):
        res = dict()
        for k, v in self.bindings.items():
            match v:
                case InjectedBind(InjectedFunction(f, args)):
                    res[k] = f.__name__
                case InjectedBind(InjectedPure(value)):
                    res[k] = str(value)
                case any:
                    res[k] = str(any)
        return res

    def build(self):
        design = self
        for k, providers in self.multi_binds.items():
            # assert k not in self.bindings,f"multi binding key overwrapping with normal binding key,{k}"
            if len(providers) == 0:
                design = design.bind_instance(**{k: set()})
            else:
                design = self._add_multi_binding(design, k, providers)
        return design

    def _ensure_provider_name(self, k, method):
        """set appropriate name for provider function to be recognized by pinject"""
        name = f"provide_{k}"
        if not method.__name__ == name:
            # there are cases where you cannot directly set __name__ attribute.
            # and sometimes pinject.inject decorator is already applied so wrapping that again is not appropriate
            # so, the solution is to first try setting __name__ and then try wrapping if failed.
            try:
                method.__name__ = name
                return method
            except AttributeError as ae:
                from loguru import logger
                logger.warning(f"somehow failed to assign new name to a provider function. trying to wrap.")

                def _wrapper(self, *args, **kwargs):
                    return method(*args, **kwargs)

                _wrapper.__name__ = name
                _wrapper.__signature__ = inspect.signature(method)
                return _wrapper
        return method

    def multi_bind_provider(self, **kwargs):
        """:cvar adds a provider to specified key so that result of calling multiple providers will be
        aggregated and provided as a list.
        """

        return self + Design(
            multi_binds={
                k: [v]
                for k, v in kwargs.items()
            }
        )

    def multi_bind_empty(self, *keys):
        """:key returns a new design which returns a [] for "key" as default value. """
        # None works as a signal to remove
        return self + Design(
            multi_binds={
                k: [None] for k in keys
            }
        )

    def _acc_multi_provider(self, providers):
        res = []
        for p in providers:
            if p is None:  # this is set by empty_multi_provider call
                res = []
            else:
                res.append(p)
        return res

    def _add_multi_binding(self, design, k, providers: list):
        from pinjected.di.util import get_class_aware_args
        # TODO use Injected's mzip.
        providers = self._acc_multi_provider(providers)
        deps = [f.dependencies() if isinstance(f, Injected) else get_class_aware_args(f) for f in providers]
        dep_set = set(chain(*deps))
        if "self" in dep_set:
            dep_set.remove("self")
        f_signature = f"multi_bind_provider_{k}({','.join(dep_set)})"
        # logger.info(f_signature)
        for ds in deps:
            for d in ds:
                assert d in dep_set

        def create_impl(tgt_providers, tgt_dependencies):
            def f_impl(**kwargs):
                # from loguru import logger

                # logger.info(f"{f_signature} called with {list(kwargs.keys())}")
                values = []
                for provider, ds in zip(tgt_providers, tgt_dependencies):
                    p_deps = {k: kwargs[k] for k in ds}
                    v = provider(**p_deps)
                    values.append(v)  # unhashable type...
                return values

            return f_impl

        new_f = create_function(f_signature, create_impl(providers, deps))
        binding = {k: new_f}
        # logger.info(binding)
        design = design.bind_provider(**binding)
        return design

    def diff(self, other):
        from pinjected.di.util import get_dict_diff
        d = get_dict_diff(self.bindings, other.bindings)
        return d

    def inspect_picklability(self):
        from pinjected.di.util import check_picklable
        from loguru import logger
        logger.info(f"checking picklability of bindings")
        check_picklable(self.bindings)
        logger.info(f"checking picklability of multi-binds")
        check_picklable(self.multi_binds)
        logger.info(f"checking picklability of modules")
        check_picklable(self.modules)
        logger.info(f"checking picklability of classes")
        check_picklable(self.classes)

    def add_modules(self, *modules):
        return self + Design(modules=list(modules))

    def add_classes(self, *classes):
        return self + Design(classes=list(classes))

    def to_vis_graph(self) -> "DIGraph":
        from pinjected.visualize_di import DIGraph
        return DIGraph(self)

    def purify(self, target: "Providable"):
        """
        given an injected, returns a minimized design which can provide the target.
        :param target:
        :return:
        """
        # ah sometimes the deps require 'session'
        # and we don't know if the session has enough bindings to provide the target.

        return self.to_graph().resolver.purified_design(target).unbind('session')


class DesignBindContext:
    def __init__(self, src: "Design", key: str):
        self.src = src
        self.key = key

    def to_class(self, cls: type):
        assert isinstance(cls, type), f"binding must be a class! got:{cls} for key:{self.src}"
        return self.src + Design({self.key: InjectedBind(Injected.bind(cls))})

    def to_instance(self, instance):
        return self.src + Design({self.key: InjectedBind(Injected.pure(instance))})

    def to_provider(self, provider):
        if isinstance(provider, Injected):
            return self.src + Design({self.key: InjectedBind(provider)})
        else:
            return self.src + Design({self.key: InjectedBind(Injected.bind(provider))})



"""
I need to define a generic structure for the followings:
- 1. Injected Function without default design
- 2. Injected Function like thing with default design
- Injected Function like thing with default design and a artifact identifier to be used as an artifact <= this can actually just be simplified as 2.

# what about the default args?
# actually, the variables canbe Injected[T] and Injected[Callable]
# but there is no way to tell that unless you actually running them.
so there are several combinations
1. Injected
2. Injected[Callable]
3. Injected with Deisgn
4. Injected[Callable] with Design

# and we need to distinguish Injected and Injected[Callable]
so, let's assume that the instance is Injected at default
and ask user to mark Injected[Callable] in the python side with annotation.
To do so we need to make Injected have some metadata... but it's not reliable.
But we can use isinstance for tagging it.
Alright.. but how do we give the optional design?
Hmm, an easy way is to add some metadata though..
Adding feature to an existing data structure is not recommended.
So I guess I need to introduce a new data structure.
"""
import loguru

import importlib
import inspect
import json
import os
from collections import defaultdict
from datetime import datetime
from pathlib import Path
from pprint import pformat
from typing import Optional, List, Dict, OrderedDict, Callable, Any

from loguru import logger
import fire
from cytoolz import memoize
from returns.maybe import Maybe, Some, maybe
from returns.result import safe, Success, Failure
from snoop import snoop

from pinjected import Injected, Design, injected_function
from pinjected.di.ast import Expr, Call, Object
from pinjected.di.injected import PartialInjectedFunction, InjectedFunction
from pinjected.di.proxiable import DelegatedVar
from pinjected.di.util import instances, providers
# from pinjected.ide_supports.create_configs import create_idea_configurations
from pinjected.helper_structure import IdeaRunConfigurations, RunnablePair, IdeaRunConfiguration
from pinjected.maybe_patch import patch_maybe
from pinjected.run_helpers.config import ConfigCreationArgs
from pinjected.helpers import find_default_design_paths
from pinjected.runnables import get_runnables, RunnableValue
from pinjected.module_var_path import ModuleVarPath, load_variable_by_module_path
from pinjected.module_inspector import ModuleVarSpec, inspect_module_for_type
from pinjected.run_config_utils_v2 import RunInjected

from pinjected.run_helpers.run_injected import run_injected

safe_getattr = safe(getattr)

patch_maybe()


@injected_function
def extract_runnables(
        default_design_path,
        logger,
        /,
        injecteds: List[ModuleVarSpec[Injected]]
):
    def extract_runnable(i: ModuleVarSpec[Injected], meta):
        match i.var, meta, default_design_path:
            case (_, Success({"default_design_path": ddp}), _):
                return RunnableValue(src=i, design_path=ddp)
            case (_, Success(), Some(ddp)):
                return RunnableValue(src=i, design_path=ddp)
            case (_, Failure(), Some(ddp)) if callable(i.var):
                return RunnableValue(src=i, design_path=ddp)
            case (DelegatedVar(), _, Some(ddp)):
                return RunnableValue(src=i, design_path=ddp)
            case (_, Failure(), Some(ddp)):
                raise ValueError(f"Cannot find default design path for {i.var_path}")
            case (_, Maybe.empty):
                raise ValueError(f"Cannot find default design path for {i.var_path}")
            case _:
                raise NotImplementedError(
                    f"Unsupported case {i, meta, default_design_path}. make sure to provide default design path.")

    results = []
    for i in injecteds:
        if isinstance(i, ModuleVarSpec):
            meta = safe(getattr)(i.var, "__runnable_metadata__")
            results.append(safe(extract_runnable)(i, meta))
    failures = [r for r in results if isinstance(r, Failure)]
    logger.warning(f"Failed to extract runnable from {failures}")
    return [r.unwrap() for r in results if isinstance(r, Success)]


@injected_function
def extract_args_for_runnable(
        logger,
        /,
        tgt: ModuleVarSpec,
        ddp: str,
        meta: dict
):
    args = None
    match tgt.var, meta:
        case (_, Success({"kind": "callable"})):
            args = ['call', tgt.var_path, ddp]
        case (_, Success({"kind": "object"})):
            args = ['get', tgt.var_path, ddp]
        case (_, Failure()) if callable(tgt.var):
            args = ['get', tgt.var_path, ddp]
        case (Injected(), Failure()):
            args = ['get', tgt.var_path, ddp]
            logger.warning(f"using get for {tgt.var_path} because it has no __runnable_metadata__")
        case (DelegatedVar(), _):
            args = ['get', tgt.var_path, ddp]
            logger.warning(f"using get for {tgt.var_path} because it has no __runnable_metadata__")
        case (_, Failure()):
            logger.info(f"skipping {tgt.var_path} because it has no __runnable_metadata__")
        case (_, Maybe.empty):
            logger.info(f"skipping {tgt.var_path} because it has no default design path.")
        case (_, Success(_meta)) if "kind" not in meta:
            args = ['get', tgt.var_path, ddp]
        case _:
            raise NotImplementedError(
                f"Unsupported case {tgt, meta, ddp}. make sure to provide default design path.")
    return args


IdeaConfigCreator = Callable[[ModuleVarSpec], List[IdeaRunConfiguration]]


@injected_function
def injected_to_idea_configs(
        runner_script_path: str,
        interpreter_path: str,
        default_design_paths: List[str],
        default_working_dir: Maybe[str],
        extract_args_for_runnable,
        logger,
        custom_idea_config_creator: IdeaConfigCreator,
        /,
        tgt: ModuleVarSpec
):
    # question is: how can we pass the override to run_injected?
    logger.info(f"using custom_idea_config_creator {custom_idea_config_creator} for {tgt}")
    name = tgt.var_path.split(".")[-1]
    # runner_script_path corresponds to the script's path which gets passed to idea.
    # so it must be the path which has run_injected command
    config_args = {
        'script_path': runner_script_path,
        'interpreter_path': interpreter_path,
        'working_dir': default_working_dir.value_or(os.getcwd())
    }

    assert isinstance(tgt, ModuleVarSpec)
    meta = safe_getattr(tgt.var, "__runnable_metadata__")
    # so we need to gather the possible design paths from the metadata too.
    ddps = []
    match meta:
        case Success({"default_design_path": ddp}):
            ddps.append(ddp)
    ddps += default_design_paths
    results = defaultdict(list)

    for ddp in ddps:
        args = extract_args_for_runnable(tgt, ddp, meta)
        # this, viz_branch should not be created by this function, but an injected function.
        if args is not None:
            ddp_name = ddp.split(".")[-1]
            config = dict(
                **config_args,
                arguments=['run_injected'] + args,
                name=f"{name}({ddp_name})"
            )
            viz_config = dict(
                **config_args,
                arguments=['run_injected', 'visualize'] + args[1:],
                name=f"{name}({ddp_name})_viz",
            )
            results[name].append(IdeaRunConfiguration(**config))
            results[name].append(IdeaRunConfiguration(**viz_config))
        else:
            logger.warning(f"skipping {tgt.var_path} because it has no __runnable_metadata__")
    if not ddps:
        logger.warning(f"no default design path provided for {tgt.var_path}")
    try:
        cfgs = custom_idea_config_creator(tgt)
        assert cfgs is not None, f"custom_idea_config_creator {custom_idea_config_creator} returned None for {tgt}. return [] if you have no custom configs."
        for configs in custom_idea_config_creator(tgt):
            results[name].append(configs)
    except Exception as e:
        logger.warning(f"Failed to create custom idea configs for {tgt} because {e}")
        raise RuntimeError(f"Failed to create custom idea configs for {tgt} because {e}") from e
    return IdeaRunConfigurations(configs=results)


notification_sounds = [
    "Basso",
    "Blow",
    "Bottle",
    "Frog",
    "Funk",
    "Glass",
    "Hero",
    "Morse",
    "Ping",
    "Pop",
    "Purr",
    "Sosumi",
    "Submarine",
    "Tink"
]


# maybe we want to distinguish the error with its complexity and pass it to gpt if required.


def run_with_kotlin(module_path: str, kotlin_zmq_address: str = None):
    d = instances()
    if kotlin_zmq_address is not None:
        d += instances(
            kotlin_zmq_address=kotlin_zmq_address
        )
    tgt: Injected = load_variable_by_module_path(module_path)
    g = d.to_graph()
    return g[tgt]


def send_kotlin_code(address: str, code: str):
    import zmq
    context = zmq.Context()
    socket = context.socket(zmq.REQ)
    socket.connect(address)
    socket.send_string(code)
    response = socket.recv_string()
    return response


def find_injecteds(
        module_path
):
    injecteds = get_runnables(module_path)
    return print(json.dumps([i.var_path for i in injecteds]))


# since this is an entrypoint, we can't use @injected_function here.


default_design = design = providers(
    # project_root=lambda module_path: Path(get_project_root(module_path)),
    # runner_script_path=lambda: self.runner_script_path or __file__,
    # interpreter_path=lambda: self.interpreter_path or sys.executable,
    # default_design_paths=lambda: find_default_design_paths(self.module_path, self.default_design_path),
    # default_working_dir=lambda project_root: maybe(lambda: self.working_dir)() | Some(
    #    str(project_root)),
    # default_design_path=lambda default_design_paths: default_design_paths[0]
) + instances(
    logger=loguru.logger,
    custom_idea_config_creator=lambda x: [],  # type ConfigCreator
    # meta_context=meta_context,
    # module_path=self.module_path,
    #
)

SANDBOX_TEMPLATE = """
from {design_path} import {design_name}
from {var_path} import {var_name}
d = {design_name} 
g = d.to_graph()
{deps}
#%%
{extras}
tgt = g[{var_name}]
"""


@maybe
def retrieve_design_path_from_injected(tgt: Injected):
    meta = safe(getattr)(tgt, '__runnable_metadata__')
    match meta:
        case {"default_design_path": design_path}:
            return design_path


def get_var_spec_from_module_path_and_name(module_path: str, var_name: str) -> ModuleVarSpec:
    tgts = get_runnables(module_path)
    name_to_tgt = {tgt.var_path.split(".")[-1]: tgt for tgt in tgts}
    tgt: ModuleVarSpec = name_to_tgt[var_name]
    return tgt


def extract_extra_codes(ast: Expr) -> ModuleVarPath:
    # TODO include the module path in the Inejcted Function
    match ast:
        case Call(
            Object(InjectedFunction(object(__original_code__=code, __name__=name, __module__=mod), _args)),
            args,
            kwargs):
            mvp = ModuleVarPath(f"{mod}.{name}")
            return mvp
        case _:
            return None


@injected_function
def make_sandbox_extra(tgt: ModuleVarSpec):
    match tgt.var:
        case DelegatedVar():
            impl_mvp = extract_extra_codes(tgt.var.eval().ast)
            import_lines = []
            impl = ""
            if impl_mvp is not None:
                import_lines += impl_mvp.depending_import_lines()
                impl = impl_mvp.definition_snippet()

            mvp = ModuleVarPath(tgt.var_path)
            import_lines += mvp.depending_import_lines()
            import_lines = '\n'.join(import_lines)
            usage = mvp.definition_snippet()
            return f"""
{import_lines}
{impl}
{usage}
"""
        case PartialInjectedFunction(InjectedFunction(object(__original_code__=usage), args)):
            return f"""
{usage}
"""
        case _:
            return ""


@injected_function
def _make_sandbox_impl(
        default_design_path: str,
        make_sandbox_extra: Callable[[Any], str],
        /,
        module_file_path: str,
        var_name: str,
):
    tgt: ModuleVarSpec = get_var_spec_from_module_path_and_name(module_file_path, var_name)
    default_design_path_parent = ".".join(default_design_path.split('.')[:-1])
    var_path_parent = ".".join(tgt.var_path.split('.')[:-1])
    # let's make a file for sandbox,
    # format datetime like 20230101_010101
    datetime_str_as_name = datetime.now().strftime("%Y%m%d_%H%M%S")
    sandbox_name = f"__sandbox__{var_name}_{datetime_str_as_name}.py"
    sandbox_path = os.path.join(os.path.dirname(module_file_path), sandbox_name)

    logger.info(f"inspecting {var_name} to make sandbox")
    injected: Injected = Injected.ensure_injected(tgt.var)
    dependencies = injected.dependencies()
    extras = make_sandbox_extra(tgt)

    with open(sandbox_path, 'w') as f:
        deps = ""
        for d in dependencies:
            deps += f"{d} = g['{d}']\n"

        f.write(SANDBOX_TEMPLATE.format(
            design_path=default_design_path_parent,
            design_name=default_design_path.split('.')[-1],
            var_path=var_path_parent,
            var_name=var_name,
            deps=deps,
            extras=extras,
        ))

    print(sandbox_path)


def make_sandbox(module_file_path, var_name):
    args = ConfigCreationArgs(
        module_path=module_file_path,
        default_design_path=None,
        runner_script_path=None,
        interpreter_path=None,
        working_dir=None,
    )
    design = args.to_design()
    g = design.to_graph()
    return g[_make_sandbox_impl](module_file_path=module_file_path, var_name=var_name)


def create_main_command(
        targets: OrderedDict[str, Injected],
        design_paths: OrderedDict[str, str],
):
    def main(target: str, design_path: Optional[str] = None):
        tgt = targets[target]
        if design_path is None:
            design_path = design_paths[list(design_paths.keys())[0]]
        design = load_variable_by_module_path(design_path)
        cmd = design.to_graph()[tgt]
        return cmd

    return main


@injected_function
def create_runnable_pair(
        main_targets: OrderedDict[str, Injected],
        default_design_paths: List[str],
        main_override_resolver,
        /,
        target: str,
        design_path: Optional[str] = None,
        overrides: str = None,
        show_graph: bool = False
) -> Optional[RunnablePair]:
    logger.info(f"creating runnable pair with {target},{design_path},{overrides}")
    logger.info(f"main targets:{pformat(main_targets.keys())},{target}")
    tgt = main_targets[target]
    if design_path is None:
        design_path = default_design_paths[0]
    main_overrides = main_override_resolver(overrides)
    design = load_variable_by_module_path(design_path) + main_overrides
    assert isinstance(design, Design), f"design at {design_path} must be Design, but got {design}"
    logger.info(f"design:{design} at {design_path}")
    pair = RunnablePair(target=tgt, design=design)
    if show_graph:
        pair.save_html()
    else:
        return pair.run()
    # cmd = design.to_graph()[tgt]
    # return cmd


def provide_module_path(logger, root_frame):
    frame_info = inspect.getframeinfo(root_frame)
    module_path = frame_info.filename
    logger.debug(f"module path:{module_path}")
    return module_path


def provide_runnables(logger, module_path) -> Dict[str, Injected]:
    tgts = get_runnables(module_path)
    name_to_tgt = {tgt.var_path.split(".")[-1]: tgt.var for tgt in tgts}
    logger.info(f"main targets:{pformat(name_to_tgt.keys())}")
    return name_to_tgt


def provide_design_paths(logger, module_path) -> OrderedDict[str, str]:
    design_paths = find_default_design_paths(module_path, None)
    design_paths = {design_path.split('.')[-1]: design_path for design_path in design_paths}
    logger.info(f"main design paths:{pformat(design_paths.keys())}")
    return design_paths


@injected_function
def main_override_resolver(query) -> Design:
    """
    :param query: can be filename with .json, .yaml.
    we can also try parsing it as json.
    :return:
    """
    import json
    from returns.pipeline import is_successful
    if isinstance(query, dict):
        return instances(**query)
    elif query is None:
        return Design()
    elif query.endswith('.json'):
        import json
        if not Path(query).exists():
            raise ValueError(f"cannot find {query} for configuration.")
        return instances(**json.load(open(query)))
    elif query.endswith('.yaml'):
        import yaml
        if not Path(query).exists():
            raise ValueError(f"cannot find {query} for configuration.")
        return instances(**yaml.load(open(query), Loader=yaml.SafeLoader))
    elif is_successful(safe(json.loads)(query)):
        return instances(**json.loads(query))
    else:
        return ModuleVarPath(query).load()


def load_variable_from_script(script_file: Path, varname: str):
    spec = importlib.util.spec_from_file_location("module.name", script_file)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return getattr(module, varname)


def parse_override_path(p) -> Design:
    if "::" in p:
        file, varname = p.split("::")
        return load_variable_by_module_path(Path(file), varname)
    else:
        load_variable_by_module_path(p)


def run_main():
    """
    inspect the caller's frame for runnable target and design path.
    delegates its execution to fire.Fire.

    I want to resolve the override, but we don't know which protocol we should use.

    get the file this is run,
    find the runnables
    find the designs
    pass it to fire.

    The protocol to override design:
    1. python file
    - /path/to/config.py::design_name
    2. python module path
    - my.package.design_name
    3. yaml or json file
    :return:
    """
    import inspect
    import fire
    from loguru import logger
    module_path = provide_module_path(logger, inspect.currentframe().f_back)
    cfg = ConfigCreationArgs(
        module_path=module_path,
    )
    # runnable: RunnablePair = (instances(
    #     root_frame=inspect.currentframe().f_back,
    #     logger=logger,
    # ) + providers(
    #     module_path=provide_module_path,
    #     main_targets=provide_runnables,
    #     main_design_paths=provide_design_paths
    # )).provide(create_runnable_pair)
    d = cfg.to_design() + instances(
        logger=logger,
    ) + providers(
        main_targets=provide_runnables,
    )
    runnable: RunnablePair = d.provide(create_runnable_pair)
    fire.Fire(runnable)


def main():
    import fire
    # maybe we should switch these commands by Injected, right?
    # we want each implementations to have design...
    # well, we can use python -m pinjected .... for these commands as well ,right?
    fire.Fire({
        # 'create_configurations': create_idea_configurations,
        'run_injected': run_injected,
        'run_injected2': RunInjected,
        'run_with_kotlin': run_with_kotlin,
        'find_injecteds': find_injecteds,
        'make_sandbox': make_sandbox,
    })


def run_idea_conf(conf: IdeaRunConfiguration, *args, **kwargs):
    pre_args = conf.arguments[1:]
    return run_injected(*pre_args, return_result=True, *args, **kwargs)


@memoize
def get_designs_from_module(module_path: Path):
    from loguru import logger
    logger.info(f"trying to import designs from {module_path}")

    def accept(name, x):
        return isinstance(x, Design) and name != "__meta_design__"

    return inspect_module_for_type(module_path, accept)


@injected_function
def get_designs_from_meta_var(
        var_path_to_file_path,
        /,
        meta: ModuleVarSpec
) -> List[ModuleVarSpec]:
    return get_designs_from_module(var_path_to_file_path(meta.var_path))


@injected_function
def var_path_to_file_path(project_root: Path, /, var_path: str) -> Path:
    module_path = '.'.join(var_path.split(".")[:-1])
    return Path(str(project_root / module_path.replace(".", os.path.sep)) + ".py")


if __name__ == '__main__':
    main()

__meta_design__ = instances(
    default_design_paths=["pinjected.run_config_utils.__meta_design__"]
)
